from asyncio import get_running_loop, iscoroutinefunction
from collections.abc import Callable, Coroutine

from draive.utils.mimic import mimic_function

__all__ = [
    "run_async",
]


def run_async[**Args, Result](
    function: Callable[Args, Result],
    /,
) -> Callable[Args, Coroutine[None, None, Result]]:
    """\
    Simple wrapper for a sync function to run in loop executor.
    The result is an async function.

    Parameters
    ----------
    function: Callable[Args, Result]
        function to be wrapped as running in loop executor

    Returns
    -------
    Callable[_Args, _Result]
        function wrapped to async using loop executor
    """

    assert not iscoroutinefunction(function), "Cannot wrap async function in executor"  # nosec: B101

    return _ExecutorWrapper(function)


class _ExecutorWrapper[**Args, Result]:
    def __init__(
        self,
        function: Callable[Args, Result],
        /,
    ) -> None:
        self._function: Callable[Args, Result] = function
        # TODO: prepare function converting kwargs to args
        # taking into account defaults

        # mimic function attributes if able
        mimic_function(function, within=self)

    async def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        assert not kwargs, "Referring to argument keys is not supported yet"  # nosec: B101

        return await get_running_loop().run_in_executor(  # pyright: ignore[reportUnknownVariableType]
            None,
            self._function,
            # not implemented by pyright - https://github.com/microsoft/pyright/discussions/5049
            *args,  # pyright: ignore[reportCallIssue]
        )



from asyncio import AbstractEventLoop, CancelledError, Event, Future, get_running_loop
from collections import deque
from collections.abc import AsyncIterator
from typing import Self

from draive.utils import freeze

__all__ = [
    "AsyncQueue",
]


class AsyncQueue[Element](AsyncIterator[Element]):
    def __init__(
        self,
        loop: AbstractEventLoop | None = None,
    ) -> None:
        self._loop: AbstractEventLoop = loop or get_running_loop()
        self._queue: deque[Element] = deque()
        self._pending: Future[Element] | None = None
        self._finished: Event = Event()

        freeze(self)

    def __del__(self) -> None:
        self.finish()

    @property
    def finished(self) -> bool:
        return self._finished.is_set()

    def enqueue(
        self,
        element: Element,
        /,
        *elements: Element,
    ) -> None:
        if self._finished.is_set():
            return  # ignore

        if pending := self._pending:
            assert not self._queue  # nosec: B101
            if not pending.done():
                pending.set_result(element)

            self._pending = None

        else:
            self._queue.append(element)

        self._queue.extend(elements)

    def cancel(self) -> None:
        if self._finished.is_set():
            return

        self._finished.set()

        while self._queue:  # clear the queue on cancel
            self._queue.popleft()

        if pending := self._pending:
            if not pending.done():
                pending.set_exception(CancelledError())

    def finish(self) -> None:
        if self._finished.is_set():
            return

        self._finished.set()

        if pending := self._pending:
            if not pending.done():
                pending.set_exception(StopAsyncIteration())

    async def wait(self) -> None:
        await self._finished.wait()

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> Element:
        if self._queue:  # check the queue, let it finish
            return self._queue.popleft()

        if self._finished.is_set():  # check if is finished
            raise StopAsyncIteration()

        # create a new future to wait for next
        assert self._pending is None, "Only a single queue iterator is supported!"  # nosec: B101
        future: Future[Element] = self._loop.create_future()
        self._pending = future

        # wait for the result
        return await future



from draive.types.audio import AudioBase64Content, AudioContent, AudioDataContent, AudioURLContent
from draive.types.errors import RateLimitError
from draive.types.frozenlist import frozenlist
from draive.types.image import ImageBase64Content, ImageContent, ImageDataContent, ImageURLContent
from draive.types.instruction import Instruction
from draive.types.json import JSON
from draive.types.lmm import (
    LMMCompletion,
    LMMCompletionChunk,
    LMMContextElement,
    LMMInput,
    LMMOutput,
    LMMOutputStream,
    LMMOutputStreamChunk,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
)
from draive.types.memory import Memory
from draive.types.multimodal import (
    MultimodalContent,
    MultimodalContentConvertible,
    MultimodalContentElement,
)
from draive.types.text import TextContent
from draive.types.tool_status import ToolCallStatus
from draive.types.video import VideoBase64Content, VideoContent, VideoDataContent, VideoURLContent

__all__ = [
    "AudioBase64Content",
    "AudioContent",
    "AudioDataContent",
    "AudioURLContent",
    "frozenlist",
    "ImageBase64Content",
    "ImageContent",
    "ImageDataContent",
    "ImageURLContent",
    "Instruction",
    "JSON",
    "LMMCompletion",
    "LMMCompletionChunk",
    "LMMContextElement",
    "LMMInput",
    "LMMOutput",
    "LMMOutputStream",
    "LMMOutputStreamChunk",
    "LMMToolRequest",
    "LMMToolRequests",
    "LMMToolResponse",
    "Memory",
    "MultimodalContent",
    "MultimodalContentConvertible",
    "MultimodalContentElement",
    "RateLimitError",
    "TextContent",
    "ToolCallStatus",
    "VideoBase64Content",
    "VideoContent",
    "VideoDataContent",
    "VideoURLContent",
]



from draive.parameters.model import DataModel

__all__ = [
    "VideoBase64Content",
    "VideoContent",
    "VideoDataContent",
    "VideoURLContent",
]


class VideoURLContent(DataModel):
    mime_type: str | None = None
    video_url: str
    video_transcription: str | None = None


class VideoBase64Content(DataModel):
    mime_type: str | None = None
    video_base64: str
    video_transcription: str | None = None


class VideoDataContent(DataModel):
    mime_type: str | None = None
    video_data: bytes
    video_transcription: str | None = None


VideoContent = VideoURLContent | VideoBase64Content | VideoDataContent



from base64 import b64encode
from collections.abc import AsyncGenerator, Sequence
from typing import Any, Literal, cast, overload

from anthropic.types import (
    ImageBlockParam,
    Message,
    MessageParam,
    TextBlock,
    TextBlockParam,
    ToolParam,
    ToolUseBlock,
)

from draive.anthropic.client import AnthropicClient
from draive.anthropic.config import AnthropicConfig
from draive.anthropic.errors import AnthropicException
from draive.metrics import ArgumentsTrace, ResultTrace, TokenUsage
from draive.parameters import DataModel, ToolSpecification
from draive.scope import ctx
from draive.types import (
    AudioBase64Content,
    AudioDataContent,
    AudioURLContent,
    ImageBase64Content,
    ImageDataContent,
    ImageURLContent,
    Instruction,
    LMMCompletion,
    LMMCompletionChunk,
    LMMContextElement,
    LMMInput,
    LMMOutput,
    LMMOutputStream,
    LMMOutputStreamChunk,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
    MultimodalContent,
    MultimodalContentElement,
    TextContent,
    VideoBase64Content,
    VideoDataContent,
    VideoURLContent,
)

__all__ = [
    "anthropic_lmm_invocation",
]


@overload
async def anthropic_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_requirement: ToolSpecification | bool | None = False,
    output: Literal["text", "json"] = "text",
    stream: Literal[True],
    **extra: Any,
) -> LMMOutputStream: ...


@overload
async def anthropic_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_requirement: ToolSpecification | bool | None = False,
    output: Literal["text", "json"] = "text",
    stream: Literal[False] = False,
    **extra: Any,
) -> LMMOutput: ...


@overload
async def anthropic_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_requirement: ToolSpecification | bool | None = False,
    output: Literal["text", "json"] = "text",
    stream: bool = False,
    **extra: Any,
) -> LMMOutputStream | LMMOutput: ...


async def anthropic_lmm_invocation(  # noqa: PLR0913
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_requirement: ToolSpecification | bool | None = False,
    output: Literal["text", "json"] = "text",
    stream: bool = False,
    **extra: Any,
) -> LMMOutputStream | LMMOutput:
    with ctx.nested(
        "anthropic_lmm_invocation",
        metrics=[
            ArgumentsTrace.of(
                instruction=instruction,
                context=context,
                tools=tools,
                tool_requirement=tool_requirement,
                output=output,
                stream=stream,
                **extra,
            ),
        ],
    ):
        ctx.log_debug("Requested Anthropic lmm")
        client: AnthropicClient = ctx.dependency(AnthropicClient)
        config: AnthropicConfig = ctx.state(AnthropicConfig).updated(**extra)
        ctx.record(config)

        messages: list[MessageParam] = [
            _convert_context_element(element=element) for element in context
        ]

        if stream:
            return ctx.stream(
                generator=_completion_stream(
                    client=client,
                    config=config,
                    instruction=Instruction.of(instruction).format(),
                    messages=messages,
                    tools=tools,
                    tool_requirement=tool_requirement,
                ),
            )

        else:
            return await _completion(
                client=client,
                config=config,
                instruction=Instruction.of(instruction).format(),
                messages=messages,
                tools=tools,
                tool_requirement=tool_requirement,
            )


def _convert_content_element(  # noqa: C901
    element: MultimodalContentElement,
) -> TextBlockParam | ImageBlockParam:
    match element:
        case TextContent() as text:
            return {
                "type": "text",
                "text": text.text,
            }

        case ImageURLContent() as image:
            # TODO: we could download the media to have data instead
            raise ValueError("Unsupported message content", element)

        case ImageBase64Content() as image:
            return {
                "type": "image",
                "source": {
                    "type": "base64",
                    "media_type": image.mime_type or "image/png",
                    "data": image.image_base64,
                },
            }

        case ImageDataContent() as image:
            return {
                "type": "image",
                "source": {
                    "type": "base64",
                    "media_type": image.mime_type or "image/png",
                    "data": b64encode(image.image_data).decode("utf-8"),
                },
            }

        case AudioURLContent():
            raise ValueError("Unsupported message content", element)

        case AudioBase64Content():
            raise ValueError("Unsupported message content", element)

        case AudioDataContent():
            raise ValueError("Unsupported message content", element)

        case VideoURLContent():
            raise ValueError("Unsupported message content", element)

        case VideoBase64Content():
            raise ValueError("Unsupported message content", element)

        case VideoDataContent():
            raise ValueError("Unsupported message content", element)

        case DataModel() as data:
            return {
                "type": "text",
                "text": data.as_json(),
            }


def _convert_context_element(
    element: LMMContextElement,
) -> MessageParam:
    match element:
        case LMMInput() as input:
            return {
                "role": "user",
                "content": [
                    _convert_content_element(element=element) for element in input.content.parts
                ],
            }

        case LMMCompletion() as completion:
            # TODO: Anthropic models generating media?
            return {
                "role": "assistant",
                "content": completion.content.as_string(),
            }

        case LMMToolRequests() as tool_requests:
            return {
                "role": "assistant",
                "content": [
                    {
                        "id": request.identifier,
                        "type": "tool_use",
                        "name": request.tool,
                        "input": request.arguments,
                    }
                    for request in tool_requests.requests
                ],
            }

        case LMMToolResponse() as tool_response:
            return {
                "role": "user",
                "content": [
                    {
                        "tool_use_id": tool_response.identifier,
                        "type": "tool_result",
                        "is_error": tool_response.error,
                        "content": [
                            _convert_content_element(element=part)
                            for part in tool_response.content.parts
                        ],
                    }
                ],
            }


async def _completion(  # noqa: PLR0913, PLR0912, C901
    *,
    client: AnthropicClient,
    config: AnthropicConfig,
    instruction: str,
    messages: list[MessageParam],
    tools: Sequence[ToolSpecification] | None,
    tool_requirement: ToolSpecification | bool | None,
) -> LMMOutput:
    completion: Message
    match tool_requirement:
        case None:
            completion = await client.completion(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[
                    ToolParam(
                        name=tool["function"]["name"],
                        description=tool["function"]["description"],
                        input_schema=cast(
                            dict[str, Any],
                            tool["function"]["parameters"],
                        ),
                    )
                    for tool in tools or []
                ],
                tool_requirement=None,
            )

        case bool(required):
            completion = await client.completion(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[
                    ToolParam(
                        name=tool["function"]["name"],
                        description=tool["function"]["description"],
                        input_schema=cast(
                            dict[str, Any],
                            tool["function"]["parameters"],
                        ),
                    )
                    for tool in tools or []
                ],
                tool_requirement=required,
            )

        case tool:
            completion = await client.completion(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[
                    ToolParam(
                        name=tool["function"]["name"],
                        description=tool["function"]["description"],
                        input_schema=cast(
                            dict[str, Any],
                            tool["function"]["parameters"],
                        ),
                    )
                    for tool in tools or []
                ],
                tool_requirement={
                    "type": "tool",
                    "name": tool["function"]["name"],
                },
            )

    ctx.record(
        TokenUsage.for_model(
            config.model,
            input_tokens=completion.usage.input_tokens or 0,
            output_tokens=completion.usage.output_tokens or 0,
        ),
    )

    message_parts: list[TextBlock] = []
    tool_calls: list[ToolUseBlock] = []
    for part in completion.content:
        match part:
            case TextBlock() as text:
                message_parts.append(text)

            case ToolUseBlock() as call:
                tool_calls.append(call)

    match completion.stop_reason:
        case "tool_use":
            if (tool_calls := tool_calls) and (tools := tools):
                ctx.record(ResultTrace.of(tool_calls))
                return LMMToolRequests(
                    requests=[
                        LMMToolRequest(
                            identifier=call.id,
                            tool=call.name,
                            arguments=cast(dict[str, Any], call.input),
                        )
                        for call in tool_calls
                    ]
                )

            else:
                raise AnthropicException("Invalid Anthropic completion", completion)

        case "end_turn":
            if message_parts:
                ctx.record(ResultTrace.of(message_parts))
                return LMMCompletion.of(
                    MultimodalContent.of(*[TextContent(text=part.text) for part in message_parts])
                )

            if (tool_calls := tool_calls) and (tools := tools):
                ctx.record(ResultTrace.of(tool_calls))
                return LMMToolRequests(
                    requests=[
                        LMMToolRequest(
                            identifier=call.id,
                            tool=call.name,
                            arguments=cast(dict[str, Any], call.input),
                        )
                        for call in tool_calls
                    ]
                )

            else:
                raise AnthropicException("Invalid Anthropic completion", completion)

        case other:
            raise AnthropicException(f"Unexpected finish reason: {other}")


async def _completion_stream(  # noqa: PLR0913
    *,
    client: AnthropicClient,
    config: AnthropicConfig,
    instruction: str,
    messages: list[MessageParam],
    tools: Sequence[ToolSpecification] | None,
    tool_requirement: ToolSpecification | bool | None,
) -> AsyncGenerator[LMMOutputStreamChunk, None]:
    ctx.log_debug("Anthropic streaming api is not supported yet, using regular response...")
    output: LMMOutput = await _completion(
        client=client,
        config=config,
        instruction=instruction,
        messages=messages,
        tools=tools,
        tool_requirement=tool_requirement,
    )

    match output:
        case LMMCompletion() as completion:
            yield LMMCompletionChunk.of(completion.content)

        case other:
            yield other



from typing import Any, Self

__all__ = [
    "ParameterValidationContext",
    "ParameterValidationError",
]

type ParameterValidationContext = tuple[str, ...]


class ParameterValidationError(Exception):
    @classmethod
    def missing(
        cls,
        context: ParameterValidationContext,
    ) -> Self:
        return cls("Validation error: Missing required parameter at: %s", context)

    @classmethod
    def invalid_type(
        cls,
        expected: Any,
        received: Any,
        context: ParameterValidationContext,
    ) -> Self:
        return cls(
            "Validation error: Invalid parameter at: %s - expected '%s' while received '%s'",
            context,
            expected,
            received,
        )

    @classmethod
    def invalid_key(
        cls,
        received: Any,
        context: ParameterValidationContext,
    ) -> Self:
        return cls(
            "Validation error: Invalid parameter key at: %s - received '%s'",
            context,
            received,
        )

    @classmethod
    def invalid_value(
        cls,
        expected: Any,
        received: Any,
        context: ParameterValidationContext,
    ) -> Self:
        return cls(
            "Validation error: Invalid parameter value at: %s - expected '%s' while received '%s'",
            context,
            expected,
            received,
        )

    @classmethod
    def invalid(
        cls,
        exception: Exception,
        context: ParameterValidationContext,
    ) -> Self:
        return cls(
            "Validation error: Invalid parameter at: %s - %s",
            context,
            exception,
        )



from asyncio import CancelledError, Event, sleep
from collections.abc import Callable

import pytest
from draive import AsyncStream, AsyncStreamTask, ctx
from pytest import raises


class FakeException(Exception):
    pass


@pytest.mark.asyncio
@ctx.wrap("test")
async def test_fails_when_streaming_fails():
    async def stream_job(send_update: Callable[[int], None]):
        raise FakeException()

    elements: int = 0
    with raises(FakeException):
        async for _ in AsyncStreamTask(job=stream_job):
            elements += 1

    assert elements == 0


@pytest.mark.asyncio
@ctx.wrap("test")
async def test_cancels_when_streaming_cancels():
    async def stream_job(send_update: Callable[[int], None]):
        await sleep(1)  # wait "forever", expecting cancellation

    elements: int = 0
    with raises(CancelledError):
        stream: AsyncStreamTask[int] = AsyncStreamTask(job=stream_job)
        ctx.cancel()
        async for _ in stream:
            elements += 1

    assert elements == 0


@pytest.mark.asyncio
@ctx.wrap("test")
async def test_ends_when_iteration_ends():
    async def stream_job(send_update: Callable[[int], None]):
        await sleep(0)  # finish without updates

    elements: int = 0
    async for _ in AsyncStreamTask(job=stream_job):
        elements += 1

    assert elements == 0


@pytest.mark.asyncio
@ctx.wrap("test")
async def test_buffers_values_when_not_reading():
    finish_event: Event = Event()

    async def stream_job(send_update: Callable[[int], None]):
        send_update(0)
        send_update(1)
        send_update(2)
        send_update(3)
        finish_event.set()

    stream: AsyncStreamTask[int] = AsyncStreamTask(job=stream_job)
    elements: int = 0
    await finish_event.wait()  # wait for all events to be sent
    async for _ in stream:
        elements += 1

    assert elements == 4


@pytest.mark.asyncio
@ctx.wrap("test")
async def test_delivers_buffer_when_streaming_fails():
    finish_event: Event = Event()

    async def stream_job(send_update: Callable[[int], None]):
        send_update(0)
        send_update(1)
        send_update(2)
        send_update(3)
        try:
            raise FakeException()
        finally:
            finish_event.set()

    elements: int = 0
    stream: AsyncStreamTask[int] = AsyncStreamTask(job=stream_job)
    await finish_event.wait()  # wait for all events to be sent
    with raises(FakeException):
        async for _ in stream:
            elements += 1

    assert elements == 4


@pytest.mark.asyncio
@ctx.wrap("test")
async def test_delivers_updates_when_sending():
    async def stream_job(send_update: Callable[[int], None]):
        for i in range(0, 10):
            send_update(i)

    elements: list[int] = []

    async for element in AsyncStreamTask(job=stream_job):
        elements.append(element)

    assert elements == list(range(0, 10))


@pytest.mark.asyncio
async def test_fails_when_sending_to_finished():
    stream: AsyncStream[int] = AsyncStream()
    stream.finish()

    with raises(RuntimeError):
        stream.send(42)


@pytest.mark.asyncio
async def test_fails_when_finishing_finished():
    stream: AsyncStream[int] = AsyncStream()
    stream.finish()

    with raises(RuntimeError):
        stream.finish()



from typing import Literal

from draive.parameters.model import DataModel

__all__ = [
    "ToolCallStatus",
]


class ToolCallStatus(DataModel):
    identifier: str
    tool: str
    status: Literal[
        "STARTED",
        "RUNNING",
        "FINISHED",
        "FAILED",
    ]
    content: DataModel | None = None



from collections.abc import Callable
from typing import Literal

from draive.lmm.invocation import LMMInvocation
from draive.parameters import DataModel, State
from draive.types import ToolCallStatus

__all__: list[str] = [
    "LMM",
    "ToolCallContext",
    "ToolStatusStream",
]


class LMM(State):
    invocation: LMMInvocation


class ToolStatusStream(State):
    send: Callable[[ToolCallStatus], None] | None = None


class ToolCallContext(State):
    call_id: str
    tool: str
    send_status: Callable[[ToolCallStatus], None]

    def report(
        self,
        status: Literal[
            "STARTED",
            "RUNNING",
            "FINISHED",
            "FAILED",
        ],
        /,
        content: DataModel | None = None,
    ) -> None:
        call_status: ToolCallStatus
        match content:
            case None:
                call_status = ToolCallStatus(
                    identifier=self.call_id,
                    tool=self.tool,
                    status=status,
                )

            case DataModel() as model:
                call_status = ToolCallStatus(
                    identifier=self.call_id,
                    tool=self.tool,
                    status=status,
                    content=model,
                )

        self.send_status(call_status)



from draive.evaluation.evaluator import (
    CaseEvaluationResult,
    CaseEvaluator,
    PreparedCaseEvaluator,
    evaluator,
)
from draive.evaluation.scenario import (
    PreparedScenarioEvaluator,
    ScenarioDefinition,
    ScenarioEvaluationResult,
    ScenarioEvaluator,
    evaluation_scenario,
)
from draive.evaluation.score import Evaluation, EvaluationScore

__all__ = [
    "CaseEvaluationResult",
    "CaseEvaluator",
    "evaluation_scenario",
    "Evaluation",
    "EvaluationScore",
    "evaluator",
    "PreparedCaseEvaluator",
    "PreparedScenarioEvaluator",
    "ScenarioDefinition",
    "ScenarioEvaluationResult",
    "ScenarioEvaluator",
]



from asyncio import Task
from collections.abc import AsyncIterator, Callable, Coroutine
from typing import Self

from draive.scope import ctx
from draive.utils import AsyncStream

__all__ = [
    "AsyncStreamTask",
]


class AsyncStreamTask[Element](AsyncIterator[Element]):
    def __init__(
        self,
        job: Callable[[Callable[[Element], None]], Coroutine[None, None, None]],
    ) -> None:
        stream: AsyncStream[Element] = AsyncStream()
        self._stream: AsyncStream[Element] = stream

        async def streaming() -> None:
            try:
                await job(stream.send)
            except Exception as exc:
                stream.finish(exc)
            else:
                stream.finish()

        self._task: Task[None] = ctx.spawn_task(streaming)

    def __del__(self) -> None:
        self._task.cancel()

    def cancel(self) -> None:
        self._task.cancel()

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> Element:
        return await self._stream.__anext__()



from asyncio import CancelledError, Task
from collections.abc import AsyncIterator
from typing import Protocol, Self, final, overload, runtime_checkable

from draive.agents.node import Agent, AgentError, AgentMessage, AgentNode
from draive.scope import ctx
from draive.utils import AsyncQueue, freeze

__all__ = [
    "AgentRunner",
    "AgentRunnerOutput",
]


@runtime_checkable
class AgentRunnerOutput(Protocol):
    @overload
    def __call__(
        self,
        error: AgentError,
        /,
    ) -> None: ...

    @overload
    def __call__(
        self,
        messages: AgentMessage,
        /,
        *_messages: AgentMessage,
    ) -> None: ...

    def __call__(
        self,
        messages: AgentMessage | AgentError,
        /,
        *_messages: AgentMessage,
    ) -> None: ...


@final
class AgentRunner:
    @classmethod
    def run(
        cls,
        agent: AgentNode,
        /,
        output: AgentRunnerOutput,
    ) -> Self:
        queue: AsyncQueue[AgentMessage] = AsyncQueue()
        return cls(
            queue=queue,
            task=ctx.spawn_subtask(
                cls._draive,
                agent,
                queue,
                output,
            ),
        )

    def __init__(
        self,
        queue: AsyncQueue[AgentMessage],
        task: Task[None],
    ) -> None:
        self._queue: AsyncQueue[AgentMessage] = queue
        self._task: Task[None] = task

        freeze(self)

    def send(
        self,
        message: AgentMessage,
        /,
    ) -> None:
        self._queue.enqueue(message)

    def cancel(self) -> None:
        self._task.cancel()

    def finish(self) -> None:
        self._queue.cancel()

    async def wait(self) -> None:
        await self._task

    async def finalize(self) -> None:
        self._queue.finish()
        await self._task

    @staticmethod
    async def _draive(
        agent: AgentNode,
        /,
        input: AsyncIterator[AgentMessage],  # noqa: A002
        output: AgentRunnerOutput,
    ) -> None:
        async with ctx.nested(agent.__str__()):
            agent_instance: Agent = agent.initialize()

            try:
                if agent.concurrent:
                    async for message in input:
                        ctx.spawn_subtask(
                            AgentRunner._handle,
                            message,
                            node=agent,
                            agent=agent_instance,
                            output=output,
                        )

                else:
                    async for message in input:
                        await AgentRunner._handle(
                            message,
                            node=agent,
                            agent=agent_instance,
                            output=output,
                        )

            except CancelledError:
                pass  # just finish when Cancelled

    @staticmethod
    async def _handle(
        message: AgentMessage,
        /,
        node: AgentNode,
        agent: Agent,
        output: AgentRunnerOutput,
    ) -> None:
        try:
            match await agent(message=message):
                case None:
                    pass  # nothing to do

                case [*results]:
                    output(*[result.updated(sender=node) for result in results])

                case result:
                    output(result.updated(sender=node))

        except CancelledError:
            pass  # ignore when Cancelled

        except Exception as exc:
            output(
                AgentError(
                    agent=node,
                    message=message,
                    cause=exc,
                )
            )



from draive.lmm.call import lmm_invocation
from draive.lmm.errors import ToolException
from draive.lmm.invocation import LMMInvocation
from draive.lmm.state import LMM, ToolCallContext, ToolStatusStream
from draive.lmm.tool import AnyTool, Tool, ToolAvailabilityCheck, tool
from draive.lmm.toolbox import Toolbox

__all__ = [
    "AnyTool",
    "ToolAvailabilityCheck",
    "lmm_invocation",
    "LMM",
    "LMMInvocation",
    "Tool",
    "Toolbox",
    "ToolCallContext",
    "ToolException",
    "ToolStatusStream",
    "tool",
]



from asyncio import AbstractEventLoop, CancelledError, TimerHandle, gather, get_running_loop
from inspect import isfunction
from typing import Protocol, cast, final, overload, runtime_checkable
from uuid import UUID, uuid4

from draive.agents.definition import AgentMessage, AgentNode
from draive.agents.node import Agent, AgentError, AgentOutput
from draive.agents.runner import AgentRunner
from draive.helpers import VolatileMemory
from draive.parameters import DataModel, ParametrizedData
from draive.scope import ctx
from draive.types import Memory, MultimodalContent, MultimodalContentConvertible, frozenlist
from draive.utils import MISSING, AsyncQueue, Missing, freeze, not_missing

__all__ = [
    "workflow",
    "AgentWorkflow",
    "AgentWorkflowInvocation",
    "AgentWorkflowOutput",
    "AgentWorkflowInput",
]


type AgentWorkflowInput = AgentMessage | AgentError
type AgentWorkflowOutput[Result: ParametrizedData | str] = (
    frozenlist[AgentMessage] | AgentMessage | Result | None
)


@runtime_checkable
class AgentWorkflowStateInitializer[AgentWorkflowState](Protocol):
    def __call__(self) -> AgentWorkflowState: ...


@runtime_checkable
class AgentWorkflowInvocation[AgentWorkflowState, AgentWorkflowResult: ParametrizedData | str](
    Protocol
):
    async def __call__(
        self,
        memory: Memory[AgentWorkflowState, AgentWorkflowState],
        input: AgentWorkflowInput,  # noqa: A002
    ) -> AgentWorkflowOutput[AgentWorkflowResult]: ...


@final
class AgentWorkflow[AgentWorkflowState, AgentWorkflowResult: ParametrizedData | str]:
    def __init__(
        self,
        node: AgentNode,
        invocation: AgentWorkflowInvocation[
            AgentWorkflowState,
            AgentWorkflowResult,
        ],
        state_initializer: AgentWorkflowStateInitializer[AgentWorkflowState],
    ) -> None:
        self.node: AgentNode = node
        self._invocation: AgentWorkflowInvocation[
            AgentWorkflowState,
            AgentWorkflowResult,
        ] = invocation
        self._state_initializer: AgentWorkflowStateInitializer[AgentWorkflowState] = (
            state_initializer
        )

        freeze(self)

    async def __call__(
        self,
        memory: Memory[AgentWorkflowState, AgentWorkflowState],
        input: AgentWorkflowInput,  # noqa: A002
    ) -> AgentWorkflowOutput[AgentWorkflowResult]:
        return await self._invocation(
            memory=memory,
            input=input,
        )

    async def run(
        self,
        input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
        state: AgentWorkflowState | None = None,
        timeout: float = 120,  # default timeout is 2 minutes
    ) -> AgentWorkflowResult:
        return await WorkflowRunner[AgentWorkflowState, AgentWorkflowResult].run(
            self,
            input=input,
            memory=VolatileMemory(state or self._state_initializer()),
            timeout=timeout,
        )

    def address(
        self,
        content: MultimodalContent | MultimodalContentConvertible,
        /,
        *_content: MultimodalContent | MultimodalContentConvertible,
        addressee: AgentNode | None = None,
    ) -> AgentMessage:
        return AgentMessage(
            identifier=uuid4(),
            sender=MISSING,
            recipient=self.node,
            addressee=addressee,
            content=MultimodalContent.of(content, *_content),
            responding=None,
        )


class PartialAgentWorkflowWrapper[AgentWorkflowState](Protocol):
    def __call__[AgentWorkflowResult: ParametrizedData | str](
        self,
        invocation: AgentWorkflowInvocation[AgentWorkflowState, AgentWorkflowResult],
    ) -> AgentWorkflow[AgentWorkflowState, AgentWorkflowResult]: ...


@overload
def workflow[AgentWorkflowState](
    node: AgentNode,
    /,
    *,
    state: AgentWorkflowStateInitializer[AgentWorkflowState],
) -> PartialAgentWorkflowWrapper[AgentWorkflowState]: ...


@overload
def workflow[AgentWorkflowState](
    *,
    name: str | None = None,
    description: str,
    state: AgentWorkflowStateInitializer[AgentWorkflowState],
) -> PartialAgentWorkflowWrapper[AgentWorkflowState]: ...


def workflow[AgentWorkflowState](
    node: AgentNode | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    state: AgentWorkflowStateInitializer[AgentWorkflowState],
) -> PartialAgentWorkflowWrapper[AgentWorkflowState]:
    assert node is None or (  # nosec: B101
        name is None and description is None
    ), "Can't specify both agent node and name/description"
    assert (  # nosec: B101
        description is not None or node is not None
    ), "Either agent node or description has to be provided"

    def wrap[AgentWorkflowResult: ParametrizedData | str](
        invocation: AgentWorkflowInvocation[AgentWorkflowState, AgentWorkflowResult],
    ) -> AgentWorkflow[AgentWorkflowState, AgentWorkflowResult]:
        assert isfunction(invocation), "workflow has to be defined from a function"  # nosec: B101

        workflow_node: AgentNode = node or AgentNode(
            name=name or invocation.__qualname__,
            description=description or "",
        )

        def initialize_agent() -> Agent:
            agent_memory: Memory[AgentWorkflowState, AgentWorkflowState] = cast(
                Memory[AgentWorkflowState, AgentWorkflowState], VolatileMemory(state())
            )

            async def agent(message: AgentMessage) -> AgentOutput:
                match await cast(  # pyright fails to properly type this function
                    AgentWorkflowInvocation[AgentWorkflowState, AgentWorkflowResult],
                    invocation,
                )(
                    memory=agent_memory,
                    input=message,
                ):
                    case None:
                        pass

                    case tuple() as messages:
                        return messages

                    case AgentMessage() as message:
                        return message

                    case result:
                        return message.respond(
                            MultimodalContent.of(
                                result if isinstance(result, DataModel) else str(result)
                            )
                        )

            return agent

        # workflow can run as an agent with limited capabilities
        workflow_node._associate(  # pyright: ignore[reportPrivateUsage]
            initialize_agent,
            concurrent=False,  # workflow can't be concurrent
        )

        return AgentWorkflow[AgentWorkflowState, AgentWorkflowResult](
            node=workflow_node,
            invocation=invocation,
            state_initializer=state,
        )

    return wrap


@runtime_checkable
class WorkflowRunnerOutput(Protocol):
    def __call__(
        self,
        messages: AgentMessage,
        /,
        *_messages: AgentMessage,
    ) -> None: ...


@runtime_checkable
class WorkflowRunnerResultOutput[WorkflowResult: ParametrizedData | str](Protocol):
    def __call__(
        self,
        result: WorkflowResult,
        /,
    ) -> None: ...


class WorkflowHistory(DataModel):
    history: frozenlist[AgentMessage]


class WorkflowRunner[WorkflowState, WorkflowResult: ParametrizedData | str]:
    @classmethod
    async def run(
        cls,
        workflow: AgentWorkflow[WorkflowState, WorkflowResult],
        /,
        input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
        memory: Memory[WorkflowState, WorkflowState],
        timeout: float,
    ) -> WorkflowResult:
        return await cls(
            workflow=workflow,
            memory=memory,
            loop=get_running_loop(),
        ).execute(
            input=input,
            timeout=timeout,
        )

    def __init__(
        self,
        workflow: AgentWorkflow[WorkflowState, WorkflowResult],
        memory: Memory[WorkflowState, WorkflowState],
        loop: AbstractEventLoop,
    ) -> None:
        self._loop: AbstractEventLoop = loop
        self._history: list[AgentMessage] = []
        self._workflow_queue: AsyncQueue[AgentMessage | AgentError] = AsyncQueue()
        self._workflow_memory: Memory[WorkflowState, WorkflowState] = memory
        self._workflow: AgentWorkflow[WorkflowState, WorkflowResult] = workflow
        self._agent_runners: dict[UUID, AgentRunner] = {}
        self._result: WorkflowResult | BaseException | Missing = MISSING

    def __del__(self) -> None:
        self.finish(result=CancelledError())

    @overload
    def send(
        self,
        error: AgentError,
        /,
    ) -> None: ...

    @overload
    def send(
        self,
        messages: AgentMessage,
        /,
        *_messages: AgentMessage,
    ) -> None: ...

    def send(
        self,
        messages: AgentMessage | AgentError,
        /,
        *_messages: AgentMessage,
    ) -> None:
        pending: list[AgentMessage | AgentError] = [messages, *_messages]
        self._history.extend([message for message in pending if isinstance(message, AgentMessage)])
        self._workflow_queue.enqueue(*pending)

    def finish(
        self,
        result: WorkflowResult | BaseException,
    ) -> None:
        if not_missing(self._result):
            return  # already finished

        ctx.log_debug("Finishing workflow with result: %s", result)

        self._result = result

        for runner in self._agent_runners.values():
            runner.finish()

        self._workflow_queue.finish()

    async def wait(self) -> None:
        await gather(
            *[runner.wait() for runner in self._agent_runners.values()],
            return_exceptions=False,
        )

        await self._workflow_queue.wait()

    async def finalize(
        self,
        result: WorkflowResult | BaseException,
    ) -> None:
        self.finish(result=result)
        await self.wait()

    async def execute(
        self,
        input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
        timeout: float,
    ) -> WorkflowResult:
        assert not self._agent_runners, "WorkflowRunner can run only once!"  # nosec: B101

        if self._workflow_queue.finished or not_missing(self._result):
            raise RuntimeError("WorkflowRunner can run only once!")

        async with ctx.nested(self._workflow.node.__str__()):

            def on_timeout() -> None:
                self.finish(result=TimeoutError())

            timeout_handle: TimerHandle = self._loop.call_later(
                delay=timeout,
                callback=on_timeout,
            )

            self.send(self._workflow.address(input))

            async for element in self._workflow_queue:
                match element:
                    case AgentMessage() as message:
                        if message.recipient.identifier == self._workflow.node.identifier:
                            await self._handle(message)

                        elif runner := self._agent_runners.get(message.recipient.identifier):
                            runner.send(message)

                        else:
                            spawned_runner: AgentRunner = AgentRunner.run(
                                message.recipient,
                                output=self.send,
                            )
                            self._agent_runners[message.recipient.identifier] = spawned_runner
                            spawned_runner.send(message)

                    case error:
                        await self._handle(error)

            await self.wait()  # wait for completion of all runners

            timeout_handle.cancel()  # cancel the timeout

            ctx.record(WorkflowHistory(history=tuple(self._history)))

            match self._result:
                case BaseException() as exc:
                    raise exc

                case Missing():
                    raise RuntimeError("Invalid workflow state")

                case result:
                    return result

    async def _handle(
        self,
        input: AgentWorkflowInput,  # noqa: A002
        /,
    ) -> None:
        try:
            match await self._workflow(
                memory=self._workflow_memory,
                input=input,
            ):
                case None:
                    pass  # nothing to do

                case [*messages]:
                    self.send(
                        *[message.updated(sender=self._workflow.node) for message in messages]
                    )

                case AgentMessage() as message:
                    self.send(message.updated(sender=self._workflow.node))

                case result:
                    self.finish(result=cast(WorkflowResult, result))

        except BaseException as exc:
            self.finish(result=exc)



from collections.abc import Callable, Coroutine
from typing import (
    Any,
    Protocol,
    cast,
    final,
    overload,
)
from uuid import uuid4

from draive.lmm.errors import ToolError, ToolException
from draive.lmm.state import ToolCallContext, ToolStatusStream
from draive.metrics import ArgumentsTrace, ResultTrace
from draive.parameters import (
    ParameterSpecification,
    ParametrizedFunction,
    ToolSpecification,
)
from draive.scope import ctx
from draive.types import MultimodalContent, MultimodalContentConvertible
from draive.utils import freeze, not_missing

__all__ = [
    "AnyTool",
    "tool",
    "Tool",
    "ToolAvailabilityCheck",
]


class ToolAvailabilityCheck(Protocol):
    def __call__(self) -> bool: ...


@final
class Tool[**Args, Result](ParametrizedFunction[Args, Coroutine[Any, Any, Result]]):
    def __init__(  # noqa: PLR0913
        self,
        /,
        name: str,
        *,
        function: Callable[Args, Coroutine[Any, Any, Result]],
        description: str | None = None,
        availability_check: ToolAvailabilityCheck | None = None,
        format_result: Callable[[Result], MultimodalContent | MultimodalContentConvertible],
        format_failure: Callable[[Exception], MultimodalContent | MultimodalContentConvertible],
        direct_result: bool = False,
    ) -> None:
        super().__init__(function=function)
        aliased_required: list[str] = []
        parameters: dict[str, ParameterSpecification] = {}
        for parameter in self._parameters.values():
            if not_missing(parameter.specification):
                parameters[parameter.aliased or parameter.name] = parameter.specification

            else:
                raise TypeError(
                    f"{function.__qualname__} can't be represented as a tool"
                    f" - argument '{parameter.name}' is missing specification."
                )

            if not (parameter.has_default or parameter.allows_missing):
                aliased_required.append(parameter.aliased or parameter.name)

        self.specification: ToolSpecification = {
            "type": "function",
            "function": {
                "name": name,
                "parameters": {
                    "type": "object",
                    "properties": parameters,
                    "required": aliased_required,
                },
                "description": description or "",
            },
        }

        self.name: str = name
        self._direct_result: bool = direct_result
        self._check_availability: ToolAvailabilityCheck = availability_check or (
            lambda: True  # available by default
        )
        self.format_result: Callable[[Result], MultimodalContent | MultimodalContentConvertible] = (
            format_result
        )
        self.format_failure: Callable[
            [Exception], MultimodalContent | MultimodalContentConvertible
        ] = format_failure

        freeze(self)

    @property
    def available(self) -> bool:
        try:
            return self._check_availability()

        except Exception:
            return False

    @property
    def requires_direct_result(self) -> bool:
        return self._direct_result

    # call from toolbox
    async def _toolbox_call(
        self,
        call_id: str,
        /,
        arguments: dict[str, Any],
    ) -> MultimodalContent:
        call_context: ToolCallContext = ToolCallContext(
            call_id=call_id,
            tool=self.name,
            send_status=ctx.state(ToolStatusStream).send or (lambda _: None),
        )
        with ctx.nested(
            self.name,
            state=[call_context],
            metrics=[ArgumentsTrace.of(**arguments)],
        ):
            call_context.report("STARTED")

            try:
                if not self.available:
                    raise ToolException(f"{self.name} is not available!")

                result: Result = await super().__call__(**arguments)  # pyright: ignore[reportCallIssue]
                ctx.record(ResultTrace.of(result))

                call_context.report("FINISHED")

                return MultimodalContent.of(self.format_result(result))

            except Exception as exc:
                call_context.report("FAILED")
                # return an error with formatted content
                raise ToolError(
                    f"Tool {self.name}[{call_id}] failed",
                    content=MultimodalContent.of(self.format_failure(exc)),
                ) from exc

    # regular call when using as a function
    async def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        with ctx.nested(
            self.name,
            state=[
                ToolCallContext(
                    call_id=uuid4().hex,
                    tool=self.name,
                    send_status=lambda _: None,
                )
            ],
            metrics=[ArgumentsTrace.of(*args, **kwargs)],
        ):
            result: Result = await super().__call__(
                *args,
                **kwargs,
            )

            ctx.record(ResultTrace.of(result))

            return result


AnyTool = Tool[Any, Any]


class ToolWrapper(Protocol):
    def __call__[**Args, Result](
        self,
        function: Callable[Args, Coroutine[Any, Any, Result]],
    ) -> Tool[Args, Result]: ...


class PartialToolWrapper[Result](Protocol):
    def __call__[**Args](
        self,
        function: Callable[Args, Coroutine[Any, Any, Result]],
    ) -> Tool[Args, Result]: ...


@overload
def tool[**Args, Result](
    function: Callable[Args, Coroutine[Any, Any, Result]],
    /,
) -> Tool[Args, Result]:
    """
    Convert a function to a tool using default parameters and no description.

    In order to adjust the arguments behavior and specification use an instance of Argument
    as a default value of any given argument with desired configuration
    for each argument individually.

    Parameters
    ----------
    function: Callable[Args, Coroutine[None, None, Result]]
        a function to be wrapped as a Tool.
    Returns
    -------
    Tool[Args, Result]
        a Tool representation of the provided function.
    """


@overload
def tool[Result](
    *,
    name: str | None = None,
    description: str | None = None,
    availability_check: ToolAvailabilityCheck | None = None,
    format_result: Callable[[Result], MultimodalContent | MultimodalContentConvertible],
    format_failure: Callable[[Exception], MultimodalContent | MultimodalContentConvertible]
    | None = None,
    direct_result: bool = False,
) -> PartialToolWrapper[Result]:
    """
    Convert a function to a tool using provided parameters.

    In order to adjust the arguments behavior and specification use an instance of Argument
    as a default value of any given argument with desired configuration
    for each argument individually.

    Parameters
    ----------
    name: str
        name to be used in a tool specification.
        Default is the name of the wrapped function.
    description: int
        description to be used in a tool specification. Allows to present the tool behavior to the
        external system.
        Default is empty.
    availability_check: ToolAvailabilityCheck
        function used to verify availability of the tool in given context. It can be used to check
        permissions or occurrence of a specific state to allow its usage.
        Provided function should raise an Exception when the tool should not be available.
        Default is always available.
    format_result: Callable[[Result], MultimodalContent]
        function converting tool result to MultimodalContent. It is used to format the result
        for model processing. Default implementation converts the result to string if needed.
    format_failure: Callable[[Exception], MultimodalContent]
        function converting tool call exception to a fallback MultimodalContent.
        Default implementation return "ERROR" string and logs the exception.
    direct_result: bool
        controls if tool result should break the ongoing processing and be the direct result of it.
        Note that during concurrent execution of multiple tools the call/result order defines
        direct result and exact behavior is not defined.
        Default is False.

    Returns
    -------
    Callable[[Function[Args, Coroutine[None, None, Result]]], Tool[Args, Result]]
        function allowing to convert other function to a Tool using provided configuration.
    """


@overload
def tool(
    *,
    name: str | None = None,
    description: str | None = None,
    availability_check: ToolAvailabilityCheck | None = None,
    format_failure: Callable[[Exception], MultimodalContent | MultimodalContentConvertible]
    | None = None,
    direct_result: bool = False,
) -> ToolWrapper:
    """
    Convert a function to a tool using provided parameters.

    In order to adjust the arguments behavior and specification use an instance of Argument
    as a default value of any given argument with desired configuration
    for each argument individually.

    Parameters
    ----------
    name: str
        name to be used in a tool specification.
        Default is the name of the wrapped function.
    description: int
        description to be used in a tool specification. Allows to present the tool behavior to the
        external system.
        Default is empty.
    availability_check: ToolAvailabilityCheck
        function used to verify availability of the tool in given context. It can be used to check
        permissions or occurrence of a specific state to allow its usage.
        Provided function should raise an Exception when the tool should not be available.
        Default is always available.
    format_result: Callable[[Result], MultimodalContent]
        function converting tool result to MultimodalContent. It is used to format the result
        for model processing. Default implementation converts the result to string if needed.
    format_failure: Callable[[Exception], MultimodalContent]
        function converting tool call exception to a fallback MultimodalContent.
        Default implementation return "ERROR" string and logs the exception.
    direct_result: bool
        controls if tool result should break the ongoing processing and be the direct result of it.
        Note that during concurrent execution of multiple tools the call/result order defines
        direct result and exact behavior is not defined.
        Default is False.

    Returns
    -------
    Callable[[Function[Args, Coroutine[None, None, Result]]], Tool[Args, Result]]
        function allowing to convert other function to a Tool using provided configuration.
    """


def tool[**Args, Result](  # noqa: PLR0913
    function: Callable[Args, Coroutine[Any, Any, Result]] | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    availability_check: ToolAvailabilityCheck | None = None,
    format_result: Callable[[Result], MultimodalContent | MultimodalContentConvertible]
    | None = None,
    format_failure: Callable[[Exception], MultimodalContent | MultimodalContentConvertible]
    | None = None,
    direct_result: bool = False,
) -> ToolWrapper | Tool[Args, Result]:
    def wrap(
        function: Callable[Args, Coroutine[Any, Any, Result]],
    ) -> Tool[Args, Result]:
        return Tool(
            name=name or function.__name__,
            description=description,
            function=function,
            availability_check=availability_check,
            format_result=format_result or _default_result_format,
            format_failure=format_failure or _default_failure_result,
            direct_result=direct_result,
        )

    if function := function:
        return wrap(function=function)

    else:
        return cast(PartialToolWrapper[Result], wrap)


def _default_result_format(result: Any) -> MultimodalContent:
    match result:
        case MultimodalContent() as content:
            return content

        case element if isinstance(element, MultimodalContentConvertible):
            return MultimodalContent.of(element)

        case other:
            return MultimodalContent.of(str(other))


def _default_failure_result(exception: Exception) -> MultimodalContent:
    ctx.log_error("Tool call failure", exception=exception)
    return MultimodalContent.of("ERROR")
